@@ -68,8 +68,8 @@ def debug_(flag,*a):
             return "\n".join((pr(y).strip("\n") for y in x))
         elif isinstance(x,bytes):
             try:
-                return unicode(x)
-            except:
+                return x.decode("utf-8")
+            except Exception:
                 return repr(x)
         elif isinstance(x,str):
             return x
@@ -103,6 +103,20 @@ def debug(*a):
 #                if v is not None:
 #                    setattr(self,f,v)
 
+class ConnEvt:
+    scope=None
+    db=None
+
+    def __init__(self):
+        self.evt = anyio.Event()
+
+    def set(self, db):
+        self.db = db
+        self.evt.set()
+
+    async def wait(self):
+        await self.evt.wait()
+
 class _db_mysql(sqlmix.db_data):
     port=3306
     def __init__(self, **kwargs):
@@ -111,15 +125,18 @@ class _db_mysql(sqlmix.db_data):
         self.DB.cursors = __import__("trio_mysql.cursors").cursors
         self.DB.paramstyle = 'format'
 
-    async def _conn(self, task_status):
+    async def _conn(self, evt):
         with anyio.CancelScope(shield=True) as sc:
+            evt.scope=sc
             async with self.DB.connect(db=self.database, host=self.host, user=self.username, password=self.password, port=self.port, **self.kwargs) as conn:
+                evt.set(conn)
                 conn._sqlmix_scope = sc
-                task_status.started(conn)
                 await anyio.sleep_forever()
 
-    async def conn(self, db):
-        return await db._tg.start(self._conn)
+    def conn(self, db):
+        evt = ConnEvt()
+        db._tg.start_soon(self._conn, evt)
+        return evt
 
 
 class _db_postgres(sqlmix.db_data):
@@ -157,6 +174,7 @@ class Db(CtxObj, sqlmix.DbPrep):
     cleaner = None
     _trace = None
     db = None
+    id_seq = 0
 
     def __init__(self,cfg=None, dbtype='mysql', _timeout=None, **kwargs):
         """\
@@ -164,8 +182,6 @@ class Db(CtxObj, sqlmix.DbPrep):
         SQL commands.
         """
 
-        self._open = set()
-
         if cfg is not None:
             try:
                 cffile = kwargs.pop('config')
@@ -214,6 +230,7 @@ class Db(CtxObj, sqlmix.DbPrep):
     @asynccontextmanager
     async def _ctx(self):
         async with anyio.create_task_group() as self._tg:
+            await self._tg.start(self._clean)
             try:
                 yield self
             finally:
@@ -236,41 +253,51 @@ class Db(CtxObj, sqlmix.DbPrep):
             r = self.db.pop()[0]
             s="OLD"
         else:
-            r = await self.DB.conn(self)
+            evt = self.DB.conn(self)
+            try:
+                await evt.wait()
+            except BaseException as e:
+                if evt.scope is not None:
+                    evt.scope.cancel()
+                # otherwise let's hope that the job won't run
+                raise
+            else:
+                r = evt.db
             s="NEW"
 
-        self._open.add(r._sqlmix_scope)
-
         debug(s)
         return r
 
     def _put_db(self,db):
-        self._open.remove(db._sqlmix_scope)
         if self.db is None or self.stopping:
-            db.close()
+            db._sqlmix_scope.cancel()
             return
         t = time()+self.timeout
         self.db.append((db,t))
         debug("BACK",getattr(db,"tid",None))
     
-    async def _clean(self):
-        self.cleaner = None
-        while self.db:
-            t = time()
-            while self.db and self.db[0][1] <= t:
-                db = self.db.pop(0)[0]
-                db.close()
-            await anyio.sleep(self.db[0][1]-t)
+    async def _clean(self, task_status):
+        with anyio.CancelScope() as self.cleaner:
+            task_status.started()
+            while True:
+                if not self.db:
+                    await anyio.sleep(self.timeout)
+                    continue
+                t = time()
+                if self.db[0][1] <= t:
+                    db = self.db.pop(0)[0]
+                    db.close()
+                    continue
+                await anyio.sleep(self.db[0][1]-t)
 
     def close(self):
+        self.stopping = True
         if self.cleaner:
             self.cleaner.cancel()
             self.cleaner = None
         while self.db:
             db = self.db.pop(0)[0]
             db._sqlmix_scope.cancel()
-        for sc in self._open:
-            sc.cancel()
         self._tg.cancel_scope.cancel()
 
     def __call__(self, job=None,retry=0):
@@ -315,7 +342,7 @@ class Db(CtxObj, sqlmix.DbPrep):
         e1 = None
         try:
             while True:
-                db = self._get_db(mtid)
+                db = self._get_db()
                 self._note(db)
                 try:
                     debug("CALL JOB",mtid)
@@ -359,24 +386,17 @@ class Db(CtxObj, sqlmix.DbPrep):
 
     async def Do(self,cmd,**kv):
         async with self() as db:
-            res = await db.Do(cmd, **kv)
-        return res
+            return await db.Do(cmd, **kv)
 
     async def DoFn(self,cmd,**kv):
         async with self() as db:
-            res = await db.DoFn(cmd, **kv)
-        return res
+            return await db.DoFn(cmd, **kv)
 
     async def DoSelect(self,cmd,**kv):
-        raise NotImplementedError("You need to call DoSelect from a transaction")
-        
-## Py3.6
-#   async def DoSelect(self,cmd,**kv):
-#       n = 0
-#       async with self() as db:
-#           async for r in db.DoSelect(cmd,**kv):
-#               yield r
-#       return n
+        n = 0
+        async with self() as db:
+            async for r in db.DoSelect(cmd,**kv):
+                yield r
 
 def _do_callback(tid,d,res):
     debug("DO_CB",tid,d,res)
@@ -389,12 +409,15 @@ class DbConn(CtxObj):
     """
     curs = None
     db = None
+    work = 0
 
     def __init__(self,pool):
         self.pool = pool
         self.committed = []
         self.rolledback = []
         self._trace = pool._trace
+        pool.id_seq += 1
+        self.id = pool.id_seq
 
     @asynccontextmanager
     async def _ctx(self):
@@ -425,7 +448,7 @@ class DbConn(CtxObj):
             self.pool._put_db(self.db)
             raise
         except BaseException:
-            db._sqlmix_scope.cancel()
+            self.db._sqlmix_scope.cancel()
             raise
         else:
             self.pool._put_db(self.db)
@@ -453,6 +476,7 @@ class DbConn(CtxObj):
                 logger.exception(proc)
 
     def close(self,reason="???"):
+        debug("CLOSE",self.id)
         if self.curs is not None:
             self.curs.close()
             self.curs = None
@@ -462,11 +486,19 @@ class DbConn(CtxObj):
             sc.cancel()
 
     async def commit(self,res=None):
+        if self.work == 0:
+            return
+        debug("COMMIT",self.id)
         await self.db.commit()
+        self.work = 0
         await self._run_committed()
 
     async def rollback(self,res=None):
+        if self.work == 0:
+            return
+        debug("ROLLBACK",self.id)
         await self.db.rollback()
+        self.work = 0
         await self._run_rolledback()
 
     async def _cursor(self, cmd, **kv):
@@ -480,6 +512,8 @@ class DbConn(CtxObj):
         return curs
         
     async def DoFn(self, cmd, **kv):
+        debug("DOFN",self.id,cmd,kv)
+        self.work += 1
         curs = await self._cursor(cmd, **kv)
 
         if hasattr(curs,'fetchone'):
@@ -511,6 +545,8 @@ class DbConn(CtxObj):
 
     async def Do(self, cmd, **kv):
         """Database-specific Do function"""
+        debug("DO",self.id,cmd,kv)
+        self.work += 1
         curs = await self._cursor(cmd, **kv)
 
         r = curs.lastrowid
@@ -521,100 +557,48 @@ class DbConn(CtxObj):
 
         if self._trace is not None:
             self._trace("Do",cmd,r)
-        if r == 0 and not '_empty' in kv:
-            raise NoData(cmd)
+        if r == 0 and not kv.get('_empty', False):
+            raise NoData(cmd, kv)
         return r
 
-# Py3.6
-#    async def DoSelect(self, cmd, **kv):
-#        """Database-specific DoSelect function"""
-#        cmd = self.db.prep(cmd, **kv)
-#
-#        try:
-#            curs = await self.cursor()
-#
-#            await curs.execute(*_cmd)
-#        except:
-#            fixup_error(_cmd)
-#            raise
-#
-#        n = 0
-#        as_dict=kv.get("_dict",None)
-#        if as_dict is True:
-#            as_dict = dict
-#        if as_dict:
-#            names = map(lambda x:x[0], curs.description)
-#
-#        try:
-#            while True:
-#                if hasattr(curs,'fetchone'):
-#                    val = await curs.fetchone()
-#                elif not curs.rows:
-#                    break
-#                else:
-#                    val = curs.rows.pop(0)
-#
-#                if ((await curs.fetchone()) is not None) if hasattr(curs,'fetchone') else curs.rows:
-#                    raise ManyData(_cmd)
-#                if as_dict:
-#                    val = as_dict(zip(names,val))
-#
-#                yield val
-#
-#        finally:
-#            if self._trace is not None:
-#                self._trace("DoSel",_cmd,val)
-#            if self.curs is None:
-#                curs.close()
-#        if n == 0 and not self.maybe_empty:
-#            raise NoData(_cmd)
+    async def DoSelect(self, cmd, **kv):
+        """Database-specific DoSelect function"""
+        debug("DOSEL",self.id,cmd,kv)
+        self.work += 1
+        curs = await self._cursor(cmd, **kv)
 
-    def DoSelect(selfi,cmd,**kv):
-        class SelectCmd(object):
-            curs = None
-            names = None
+        n = 0
+        as_dict=kv.get("_dict",None)
+        if as_dict is True:
+            as_dict = dict
+        if as_dict:
+            names = list(map(lambda x:x[0], curs.description))
 
-            def __init__(self,cmd,**k):
-                self.cmd = cmd
-                self.k = k
-
-                self.n = 0
-                self.as_dict=k.pop("_dict",False)
-                self.maybe_empty=k.pop("_empty",False)
-                if self.as_dict is True:
-                    self.as_dict = dict
-
-            def __aiter__(self):
-                return self
-
-            async def __anext__(self):
-                if self.curs is None:
-                    self.curs = curs = await selfi._cursor(cmd, **kv)
-
-                    if self.as_dict:
-                        self.names = list(map(lambda x:x[0], curs.description))
-                else:
-                    curs = self.curs
-                
+        try:
+            while True:
                 if hasattr(curs,'fetchone'):
                     val = await curs.fetchone()
+                    if val is None:
+                        break
                 elif not curs.rows:
-                    val = None
+                    break
                 else:
                     val = curs.rows.pop(0)
 
-                if val is None:
-                    if self.n == 0 and not self.maybe_empty:
-                        raise NoData(self.cmd)
-                    raise StopAsyncIteration
+                if as_dict:
+                    val = as_dict(zip(names,val))
 
-                if self.as_dict:
-                    val = self.as_dict(zip(self.names,val))
+                n += 1
+                yield val
 
-                self.n += 1
-                return val
-
-        return SelectCmd(cmd,**kv)
+        finally:
+            if self._trace is not None:
+                self._trace("DoSel",_cmd,val)
+            if self.curs is None:
+                with anyio.move_on_after(3, shield=True):
+                    await curs.aclose()
+        if n == 0 and not kv.get('_empty', False):
+            raise NoData(cmd, kv)
 
     Do.__doc__ = sqlmix.Db.Do.__doc__ + "\nReturns a Future.\n"
     DoFn.__doc__ = sqlmix.Db.DoFn.__doc__ + "\nReturns a Future.\n"
